"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const tag_expressions_1 = __importDefault(require("@cucumber/tag-expressions"));
const uuid_1 = require("uuid");
const assertions_1 = require("./assertions");
const data_table_1 = __importDefault(require("./data_table"));
const registry_1 = require("./registry");
const ast_helpers_1 = require("./ast-helpers");
const constants_1 = require("./constants");
const environment_helpers_1 = require("./environment-helpers");
const type_guards_1 = require("./type-guards");
/**
 * From messages.TestStepFinished.TestStepResult.Status.
 */
const Status = {
    Unknown: "UNKNOWN",
    Passed: "PASSED",
    Skipped: "SKIPPED",
    Pending: "PENDING",
    Undefined: "UNDEFINED",
    Ambiguous: "AMBIGUOUS",
    Failed: "FAILED",
};
const sourceReference = {
    uri: "not available",
    location: { line: 0 },
};
function retrieveInternalProperties() {
    return Cypress.env(constants_1.INTERNAL_PROPERTY_NAME);
}
function findPickleById(context, astId) {
    return (0, assertions_1.assertAndReturn)(context.pickles.find((pickle) => pickle.astNodeIds && pickle.astNodeIds.includes(astId)), `Expected to find a pickle associated with id = ${astId}`);
}
function collectExampleIds(examples) {
    return examples
        .map((examples) => {
        return (0, assertions_1.assertAndReturn)(examples.tableBody, "Expected to find a table body").map((row) => (0, assertions_1.assertAndReturn)(row.id, "Expected table row to have an id"));
    })
        .reduce((acum, el) => acum.concat(el), []);
}
function createTimestamp() {
    const now = new Date().getTime();
    const seconds = Math.floor(now / 1000);
    const nanos = (now - seconds * 1000) * 1000000;
    return {
        seconds,
        nanos,
    };
}
function duration(start, end) {
    return {
        seconds: end.seconds - start.seconds,
        nanos: end.nanos - start.nanos,
    };
}
function createFeature(context, feature) {
    describe(feature.name || "<unamed feature>", () => {
        if (feature.children) {
            for (const child of feature.children) {
                if (child.scenario) {
                    createScenario(context, child.scenario);
                }
                else if (child.rule) {
                    createRule(context, child.rule);
                }
            }
        }
    });
}
function createRule(context, rule) {
    var _a;
    const picklesWithinRule = (_a = rule.children) === null || _a === void 0 ? void 0 : _a.map((child) => child.scenario).filter(type_guards_1.notNull).flatMap((scenario) => {
        if (scenario.examples) {
            return collectExampleIds(scenario.examples).map((exampleId) => {
                return findPickleById(context, exampleId);
            });
        }
        else {
            const scenarioId = (0, assertions_1.assertAndReturn)(scenario.id, "Expected scenario to have an id");
            return findPickleById(context, scenarioId);
        }
    });
    if (picklesWithinRule) {
        if (context.omitFiltered) {
            const matches = picklesWithinRule.filter((pickle) => context.testFilter.evaluate((0, ast_helpers_1.collectTagNames)(pickle.tags)));
            if (matches.length === 0) {
                return;
            }
        }
    }
    describe(rule.name || "<unamed rule>", () => {
        if (rule.children) {
            for (const child of rule.children) {
                if (child.scenario) {
                    createScenario(context, child.scenario);
                }
            }
        }
    });
}
function createWeakCache(mapper) {
    return {
        cache: new WeakMap(),
        get(key) {
            const cacheHit = this.cache.get(key);
            if (cacheHit) {
                return cacheHit;
            }
            const value = mapper(key);
            this.cache.set(key, value);
            return value;
        },
    };
}
const gherkinDocumentsAstIdMaps = createWeakCache((key) => {
    const astIdMap = new Map();
    for (const node of (0, ast_helpers_1.traverseGherkinDocument)(key)) {
        if ("id" in node && node.id) {
            astIdMap.set(node.id, node);
        }
    }
    return astIdMap;
});
function createScenario(context, scenario) {
    if (scenario.examples) {
        const exampleIds = collectExampleIds(scenario.examples);
        for (let i = 0; i < exampleIds.length; i++) {
            const exampleId = exampleIds[i];
            const pickle = findPickleById(context, exampleId);
            const baseName = (0, assertions_1.assertAndReturn)(pickle.name, "Expected pickle to have a name");
            const exampleName = `${baseName} (example #${i + 1})`;
            createPickle(context, Object.assign(Object.assign({}, scenario), { name: exampleName }), pickle);
        }
    }
    else {
        const scenarioId = (0, assertions_1.assertAndReturn)(scenario.id, "Expected scenario to have an id");
        const pickle = findPickleById(context, scenarioId);
        createPickle(context, scenario, pickle);
    }
}
function createPickle(context, scenario, pickle) {
    var _a;
    const { registry, gherkinDocument, pickles, testFilter, messages } = context;
    const testCaseId = (0, uuid_1.v4)();
    const pickleSteps = (_a = pickle.steps) !== null && _a !== void 0 ? _a : [];
    const scenarioName = scenario.name || "<unamed scenario>";
    const tags = (0, ast_helpers_1.collectTagNames)(pickle.tags);
    const beforeHooks = registry.resolveBeforeHooks(tags);
    const afterHooks = registry.resolveAfterHooks(tags);
    const definitionIds = pickleSteps.map(() => (0, uuid_1.v4)());
    const steps = [
        ...beforeHooks.map((hook) => ({ hook })),
        ...pickleSteps.map((pickleStep) => ({ pickleStep })),
        ...afterHooks.map((hook) => ({ hook })),
    ];
    for (const id of definitionIds) {
        messages.stack.push({
            stepDefinition: {
                id,
                pattern: {
                    source: "a step",
                    type: "CUCUMBER_EXPRESSION",
                },
                sourceReference,
            },
        });
    }
    const testSteps = [];
    for (const beforeHook of beforeHooks) {
        testSteps.push({
            id: beforeHook.id,
            hookId: beforeHook.id,
        });
    }
    for (let i = 0; i < pickleSteps.length; i++) {
        const step = pickleSteps[i];
        testSteps.push({
            id: step.id,
            pickleStepId: step.id,
            stepDefinitionIds: [definitionIds[i]],
        });
    }
    for (const afterHook of afterHooks) {
        testSteps.push({
            id: afterHook.id,
            hookId: afterHook.id,
        });
    }
    messages.stack.push({
        testCase: {
            id: testCaseId,
            pickleId: pickle.id,
            testSteps,
        },
    });
    const astIdMap = gherkinDocumentsAstIdMaps.get(gherkinDocument);
    if (!testFilter.evaluate(tags) || tags.includes("@skip")) {
        if (!context.omitFiltered) {
            it.skip(scenarioName);
        }
        return;
    }
    let attempt = 0;
    const internalProperties = {
        pickle,
        testCaseStartedId: (0, uuid_1.v4)(),
        allSteps: steps,
        remainingSteps: [...steps],
    };
    const env = { [constants_1.INTERNAL_PROPERTY_NAME]: internalProperties };
    it(scenarioName, { env }, function () {
        var _a, _b, _c, _d, _e;
        const { remainingSteps, testCaseStartedId } = retrieveInternalProperties();
        (0, registry_1.assignRegistry)(registry);
        messages.stack.push({
            testCaseStarted: {
                id: testCaseStartedId,
                testCaseId,
                attempt: attempt++,
                timestamp: createTimestamp(),
            },
        });
        window.testState = {
            gherkinDocument,
            pickles,
            pickle,
        };
        for (const step of steps) {
            if (step.hook) {
                const hook = step.hook;
                cy.then(() => {
                    Cypress.log({
                        name: "step",
                        displayName: hook.keyword,
                        message: "",
                    });
                    const start = createTimestamp();
                    messages.stack.push({
                        testStepStarted: {
                            testStepId: hook.id,
                            testCaseStartedId,
                            timestamp: start,
                        },
                    });
                    if (messages.enabled) {
                        cy.task(constants_1.TASK_TEST_STEP_STARTED, hook.id, { log: false });
                    }
                    return cy.wrap(start, { log: false });
                })
                    .then((start) => {
                    registry.runHook(this, hook);
                    return cy.wrap(start, { log: false });
                })
                    .then((start) => {
                    const end = createTimestamp();
                    messages.stack.push({
                        testStepFinished: {
                            testStepId: hook.id,
                            testCaseStartedId,
                            testStepResult: {
                                status: Status.Passed,
                                duration: duration(start, end),
                            },
                            timestamp: end,
                        },
                    });
                    remainingSteps.shift();
                });
            }
            else if (step.pickleStep) {
                const pickleStep = step.pickleStep;
                const text = (0, assertions_1.assertAndReturn)(pickleStep.text, "Expected pickle step to have a text");
                const scenarioStep = (0, assertions_1.assertAndReturn)(astIdMap.get((0, assertions_1.assertAndReturn)((_a = pickleStep.astNodeIds) === null || _a === void 0 ? void 0 : _a[0], "Expected to find at least one astNodeId")), `Expected to find scenario step associated with id = ${(_b = pickleStep.astNodeIds) === null || _b === void 0 ? void 0 : _b[0]}`);
                cy.then(() => {
                    Cypress.log({
                        name: "step",
                        displayName: (0, assertions_1.assertAndReturn)("keyword" in scenarioStep && scenarioStep.keyword, "Expected to find a keyword in the scenario step"),
                        message: text,
                    });
                });
                const argument = ((_c = pickleStep.argument) === null || _c === void 0 ? void 0 : _c.dataTable)
                    ? new data_table_1.default(pickleStep.argument.dataTable)
                    : ((_e = (_d = pickleStep.argument) === null || _d === void 0 ? void 0 : _d.docString) === null || _e === void 0 ? void 0 : _e.content)
                        ? pickleStep.argument.docString.content
                        : undefined;
                cy.then(() => {
                    internalProperties.currentStep = { pickleStep };
                    const start = createTimestamp();
                    messages.stack.push({
                        testStepStarted: {
                            testStepId: pickleStep.id,
                            testCaseStartedId,
                            timestamp: start,
                        },
                    });
                    if (messages.enabled) {
                        cy.task(constants_1.TASK_TEST_STEP_STARTED, pickleStep.id, { log: false });
                    }
                    return cy.wrap(start, { log: false });
                })
                    .then((start) => {
                    const ensureChain = (value) => Cypress.isCy(value) ? value : cy.wrap(value, { log: false });
                    return ensureChain(registry.runStepDefininition(this, text, argument)).then((result) => {
                        return {
                            start,
                            result,
                        };
                    });
                })
                    .then(({ start, result }) => {
                    var _a, _b, _c;
                    const end = createTimestamp();
                    if (result === "pending") {
                        messages.stack.push({
                            testStepFinished: {
                                testStepId: pickleStep.id,
                                testCaseStartedId,
                                testStepResult: {
                                    status: Status.Pending,
                                    duration: duration(start, end),
                                },
                                timestamp: end,
                            },
                        });
                        remainingSteps.shift();
                        for (const skippedStep of remainingSteps) {
                            const testStepId = (0, assertions_1.assertAndReturn)((_b = (_a = skippedStep.hook) === null || _a === void 0 ? void 0 : _a.id) !== null && _b !== void 0 ? _b : (_c = skippedStep.pickleStep) === null || _c === void 0 ? void 0 : _c.id, "Expected a step to either be a hook or a pickleStep");
                            messages.stack.push({
                                testStepStarted: {
                                    testStepId,
                                    testCaseStartedId,
                                },
                            });
                            messages.stack.push({
                                testStepFinished: {
                                    testStepId,
                                    testCaseStartedId,
                                    testStepResult: {
                                        status: Status.Skipped,
                                    },
                                },
                            });
                        }
                        for (let i = 0, count = remainingSteps.length; i < count; i++) {
                            remainingSteps.pop();
                        }
                        this.skip();
                    }
                    else {
                        messages.stack.push({
                            testStepFinished: {
                                testStepId: pickleStep.id,
                                testCaseStartedId,
                                testStepResult: {
                                    status: Status.Passed,
                                    duration: duration(start, end),
                                },
                                timestamp: end,
                            },
                        });
                        remainingSteps.shift();
                    }
                });
            }
        }
    });
}
function collectTagNamesFromGherkinDocument(gherkinDocument) {
    const tagNames = [];
    for (const node of (0, ast_helpers_1.traverseGherkinDocument)(gherkinDocument)) {
        if ("tags" in node) {
            tagNames.push(...(0, ast_helpers_1.collectTagNames)(node.tags));
        }
    }
    return tagNames;
}
function createTests(registry, source, gherkinDocument, pickles, messagesEnabled, omitFiltered) {
    const noopNode = { evaluate: () => true };
    const environmentTags = (0, environment_helpers_1.getTags)(Cypress.env());
    const messages = [];
    messages.push({
        source: { data: source, uri: gherkinDocument.uri },
    });
    messages.push({
        gherkinDocument: Object.assign({}, gherkinDocument),
    });
    for (const pickle of pickles) {
        messages.push({
            pickle,
        });
    }
    for (const hook of [...registry.beforeHooks, ...registry.afterHooks]) {
        messages.push({
            hook: {
                id: hook.id,
                sourceReference,
            },
        });
    }
    const tagsInDocument = collectTagNamesFromGherkinDocument(gherkinDocument);
    const testFilter = tagsInDocument.includes("@only") || tagsInDocument.includes("@focus")
        ? (0, tag_expressions_1.default)("@only or @focus")
        : environmentTags
            ? (0, tag_expressions_1.default)(environmentTags)
            : noopNode;
    if (gherkinDocument.feature) {
        createFeature({
            registry,
            gherkinDocument,
            pickles,
            testFilter,
            omitFiltered,
            messages: {
                enabled: messagesEnabled,
                stack: messages,
            },
        }, gherkinDocument.feature);
    }
    const isHooksAttached = globalThis[constants_1.INTERNAL_PROPERTY_NAME];
    if (isHooksAttached) {
        return;
    }
    else {
        globalThis[constants_1.INTERNAL_PROPERTY_NAME] = true;
    }
    afterEach(function () {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l;
        (0, registry_1.freeRegistry)();
        const properties = retrieveInternalProperties();
        const { testCaseStartedId, remainingSteps } = properties;
        if (remainingSteps.length > 0 &&
            ((_a = this.currentTest) === null || _a === void 0 ? void 0 : _a.state) !== "pending") {
            const error = (0, assertions_1.assertAndReturn)((_c = (_b = this.currentTest) === null || _b === void 0 ? void 0 : _b.err) === null || _c === void 0 ? void 0 : _c.message, "Expected to find an error message");
            if (constants_1.HOOK_FAILURE_EXPR.test(error)) {
                return;
            }
            const failedStep = (0, assertions_1.assertAndReturn)(remainingSteps.shift(), "Expected there to be a remaining step");
            const testStepId = (0, assertions_1.assertAndReturn)((_e = (_d = failedStep.hook) === null || _d === void 0 ? void 0 : _d.id) !== null && _e !== void 0 ? _e : (_f = failedStep.pickleStep) === null || _f === void 0 ? void 0 : _f.id, "Expected a step to either be a hook or a pickleStep");
            const failedTestStepFinished = error.includes("Step implementation missing")
                ? {
                    testStepFinished: {
                        testStepId,
                        testCaseStartedId,
                        testStepResult: {
                            status: Status.Undefined,
                        },
                    },
                }
                : {
                    testStepFinished: {
                        testStepId,
                        testCaseStartedId,
                        testStepResult: {
                            status: Status.Failed,
                            message: (_h = (_g = this.currentTest) === null || _g === void 0 ? void 0 : _g.err) === null || _h === void 0 ? void 0 : _h.message,
                        },
                        timestamp: createTimestamp(),
                    },
                };
            messages.push(failedTestStepFinished);
            for (const skippedStep of remainingSteps) {
                const testStepId = (0, assertions_1.assertAndReturn)((_k = (_j = skippedStep.hook) === null || _j === void 0 ? void 0 : _j.id) !== null && _k !== void 0 ? _k : (_l = skippedStep.pickleStep) === null || _l === void 0 ? void 0 : _l.id, "Expected a step to either be a hook or a pickleStep");
                messages.push({
                    testStepStarted: {
                        testStepId,
                        testCaseStartedId,
                    },
                });
                messages.push({
                    testStepFinished: {
                        testStepId,
                        testCaseStartedId,
                        testStepResult: {
                            status: Status.Skipped,
                        },
                    },
                });
            }
        }
        messages.push({
            testCaseFinished: {
                testCaseStartedId,
                timestamp: createTimestamp(),
            },
        });
        /**
         * Repopulate internal properties in case previous test is retried.
         */
        properties.testCaseStartedId = (0, uuid_1.v4)();
        properties.remainingSteps = properties.allSteps;
    });
    after(function () {
        if (messagesEnabled) {
            cy.task(constants_1.TASK_APPEND_MESSAGES, messages, { log: false });
        }
    });
}
exports.default = createTests;
